@@ -1,11 +1,12 @@
 /**
- * LIBOMEMO 0.7.0
+ * LIBOMEMO 0.8.1
  */
 
 
 #pragma once
 
 #include <inttypes.h>
+#include <stdbool.h>
 
 #include <glib.h>
 
@@ -88,16 +89,48 @@ typedef struct omemo_crypto_provider {
 #define OMEMO_LOG_INFO    3
 #define OMEMO_LOG_DEBUG   4
 
-#define OMEMO_SUCCESS                      0
-#define OMEMO_ERR                     -10000
-#define OMEMO_ERR_NOMEM               -10001
-#define OMEMO_ERR_NULL                -10002
-#define OMEMO_ERR_CRYPTO              -10010
-#define OMEMO_ERR_AUTH_FAIL           -10020
-#define OMEMO_ERR_UNSUPPORTED_KEY_LEN -10030
-#define OMEMO_ERR_STORAGE             -10100
-#define OMEMO_ERR_MALFORMED_BUNDLE    -11000
-#define OMEMO_ERR_MALFORMED_XML       -12000
+#define OMEMO_SUCCESS                                                  0
+#define OMEMO_ERR                                                 -10000
+#define OMEMO_ERR_NOMEM                                           -10001
+#define OMEMO_ERR_NULL                                            -10002
+#define OMEMO_ERR_CRYPTO                                          -10010
+#define OMEMO_ERR_AUTH_FAIL                                       -10020
+#define OMEMO_ERR_UNSUPPORTED_KEY_LEN                             -10030
+#define OMEMO_ERR_STORAGE                                         -10100
+
+// the errors below were initially all equal to the first one
+#define OMEMO_ERR_MALFORMED_BUNDLE                                -11000
+#define OMEMO_ERR_MALFORMED_BUNDLE_NO_ITEMS_ELEM                  -11001
+#define OMEMO_ERR_MALFORMED_BUNDLE_NO_NODE_ATTR                   -11002
+#define OMEMO_ERR_MALFORMED_BUNDLE_NO_ITEM_ELEM                   -11003
+#define OMEMO_ERR_MALFORMED_BUNDLE_NO_BUNDLE_ELEM                 -11004
+#define OMEMO_ERR_MALFORMED_BUNDLE_NO_SPK_ELEM                    -11005
+#define OMEMO_ERR_MALFORMED_BUNDLE_NO_SPK_DATA                    -11006
+#define OMEMO_ERR_MALFORMED_BUNDLE_NO_SPK_ID_ATTR                 -11007
+#define OMEMO_ERR_MALFORMED_BUNDLE_NO_SIG_ELEM                    -11008
+#define OMEMO_ERR_MALFORMED_BUNDLE_NO_SIG_DATA                    -11009
+#define OMEMO_ERR_MALFORMED_BUNDLE_NO_IK_ELEM                     -11010
+#define OMEMO_ERR_MALFORMED_BUNDLE_NO_IK_DATA                     -11011
+#define OMEMO_ERR_MALFORMED_BUNDLE_NO_PREKEYS_ELEM                -11012
+#define OMEMO_ERR_MALFORMED_BUNDLE_NO_PREKEY_ELEM                 -11013
+#define OMEMO_ERR_MALFORMED_BUNDLE_NO_PREKEY_DATA                 -11014
+#define OMEMO_ERR_MALFORMED_BUNDLE_NO_PREKEY_ID_ATTR              -11015
+
+// the errors below were also initially all equal to the first one
+#define OMEMO_ERR_MALFORMED_XML                                   -12000
+#define OMEMO_ERR_MALFORMED_DEVICELIST_NO_ITEMS_ELEM              -12101
+#define OMEMO_ERR_MALFORMED_DEVICELIST_NO_ITEM_ELEM               -12102
+#define OMEMO_ERR_MALFORMED_DEVICELIST_NO_LIST_ELEM               -12103
+#define OMEMO_ERR_MALFORMED_DEVICELIST_NO_DEVICE_ID_ATTR          -12104
+#define OMEMO_ERR_MALFORMED_OUTGOING_MESSAGE_NO_BODY_ELEM         -12201
+#define OMEMO_ERR_MALFORMED_OUTGOING_MESSAGE_NO_BODY_DATA         -12202
+#define OMEMO_ERR_MALFORMED_INCOMING_MESSAGE_NO_BODY_ELEM         -12301
+#define OMEMO_ERR_MALFORMED_INCOMING_MESSAGE_NO_ENCRYPTED_ELEM    -12302
+#define OMEMO_ERR_MALFORMED_INCOMING_MESSAGE_NO_HEADER_ELEM       -12303
+#define OMEMO_ERR_MALFORMED_INCOMING_MESSAGE_NO_PAYLOAD_DATA      -12304
+#define OMEMO_ERR_MALFORMED_INCOMING_MESSAGE_NO_IV_ELEM           -12305
+#define OMEMO_ERR_MALFORMED_INCOMING_MESSAGE_NO_IV_DATA           -12306
+#define OMEMO_ERR_MALFORMED_INCOMING_MESSAGE_NO_KEY_DATA          -12307
 
 #define OMEMO_ADD_MSG_NONE 0
 #define OMEMO_ADD_MSG_BODY 1
@@ -427,6 +460,7 @@ int omemo_message_prepare_encryption(char * outgoing_message, uint32_t sender_de
  */
 const uint8_t * omemo_message_get_key(omemo_message * msg_p);
 
+
 /**
  * Gets the length of the symmetric key and appended tag.
  * Again, this only makes sense on an outgoing message.
@@ -451,6 +485,20 @@ size_t omemo_message_get_key_len(omemo_message * msg_p);
  */
 int omemo_message_add_recipient(omemo_message * msg_p, uint32_t device_id, const uint8_t * encrypted_key_p, size_t key_len);
 
+/**
+ * Add the encrypted symmetric key for a specific device id to the message with the prekey attribute.
+ * This notifies the recipient device of the fact that this is the very first message of the session,
+ * which therefore still has to be estalished on its side. 
+ * Only makes sense on outgoing messages.
+ *
+ * @param msg_p Pointer to the message to add to.
+ * @param device_id The recipient device ID.
+ * @param encrypted_key_p The encrypted key data.
+ * @param key_len Length of the encrypted key data.
+ * @return 0 on success, negative on error.
+ */
+int omemo_message_add_recipient_w_prekey(omemo_message * msg_p, uint32_t device_id, const uint8_t * encrypted_key_p, size_t key_len);
+
 /**
  * After all recipients have been added, this function can be used to export the resulting <message> stanza.
  * Also adds a <store> hint.
@@ -534,6 +582,16 @@ char * omemo_message_get_recipient_name_bare(omemo_message * msg_p);
  */
 int omemo_message_get_encrypted_key(omemo_message * msg_p, uint32_t own_device_id, uint8_t ** key_pp, size_t * key_len_p );
 
+/**
+ * Returns whether the key with the speicified device ID is a prekey.
+ * 
+ * @param msg_p Pointer to the message.
+ * @param recipient_device_id  The device ID to check.
+ * @param is_prekey_p Will be set to true if the key is a prekey, false otherwise.
+ * @return int 0 on success, negative on error.
+ */
+int omemo_message_is_encrypted_key_prekey(omemo_message * msg_p, uint32_t recipient_device_id, bool * is_prekey_p);
+
 /**
  * Using the decrypted symmetric key, this method decrypts the payload and exports the original <message> stanza.
  *
